
#ifndef PLS_TBB_LIKE_TASK_H
#define PLS_TBB_LIKE_TASK_H

#include "abstract_task.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            class tbb_task;
            class tbb_sub_task {
                friend class tbb_task;

                std::atomic_uint32_t ref_count_;
                tbb_sub_task* parent_;
                tbb_task* tbb_task_;

            public:
                explicit tbb_sub_task(tbb_sub_task* parent, tbb_task* tbb_task);
                ~tbb_sub_task();
                void execute();

            protected:
                virtual void execute_internal() = 0;
                // SubClass Implementations:
                // Do Work
                // |-- Spawn Sub Task
                // |-- Spawn Sub task
                // Do Work
                // |-- Wait For All
                // Do  Work
                // |-- Spawn Sub Task

                // Currently required to construct child...
                // TODO: Allocate child with custom new(...) on stack
                tbb_sub_task* parent() { return parent_; }
                tbb_task* tbb_task() { return tbb_task_; }

                void wait_for_all();
            private:
                tbb_sub_task* get_local_task();
            };

            class tbb_task: public abstract_task {
                friend class tbb_sub_task;
                tbb_sub_task* root_task_;
                // TODO: hold stuff for double ended sub-task queue

                explicit tbb_task(tbb_sub_task* root_task):
                    abstract_task{0, 0},
                    root_task_{root_task} {};

                void execute() override {
                    root_task_->execute();
                }

                bool internal_stealing(abstract_task* other_task) override {
                    auto cast_other_task = reinterpret_cast<tbb_task*>(other_task);
                    // TODO: Try to steal from the other sub-task queue
                    return false;
                }

                bool split_task() override {
                    // TODO: Take an internal task and create a new tbb task from it
                    return false;
                }
            };
        }
    }
}

#endif //PLS_TBB_LIKE_TASK_H
